
/*	
	This file is part of miSCellaneous, a program library for SuperCollider 3
	
	Created: 2011-08-18, version 0.4_beta
	Copyright (C) 2009, 2010, 2011 Daniel Mayer 
	Email: 	daniel-mayer@email.de
	URL:		www.daniel-mayer.at

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

+VarGui	{
	
	checkSingleSynthInput { |synth|
		case 	
			{ [Symbol, String].any(synth.isKindOf(_)) }
				{ 
					SynthDescLib.global[synth.asSymbol].isNil.if {
						SimpleInitError("Synth given as " ++ synth.class.asString ++ " " ++ synth.asString ++ " not contained " ++ 
								"in global SynthDescLib (forgot SynthDef.add, former .memStore ?)").class_(VarGui).throw;
					}
				}	
			{ [Synth, Integer].any(synth.isKindOf(_)).not }
				{ SimpleInitError("Single synth input item must be synthdef (given as Symbol or String) or " ++ 
					"Synth (directly or as nodeID integer)").class_(VarGui).throw };
		^this
	}
			
	// not a complete input check, but avoid some nearby discrepances
		
	checkArgs {|varCtr, synthCtr, stream, synth, varEnvirGroups, envir, synthCtrGroups, assumePlaying, assumeRunning, assumeEnded|
		var j = 0, vars, synthCtrs, envirSet, synthStateVars, synthStateStrings, synthInputDup = false, result;
		
		// sythCtr input must grouped per synth, varCtr input can be grouped or not

		this.varCtr = case
			{ varCtr.isNil }{ [] }
			{ varCtr.first.isKindOf(SequenceableCollection) }{ varCtr.flatten } 
			{ true }{ varCtr };
			 
		this.synthCtr = case
			{ synthCtr.isNil }{ [] }
			{ synthCtr.first.isKindOf(SequenceableCollection) }{ synthCtr.flatten } 
			{ true }{ synthCtr };
			
		vars = this.varCtr.select {|x,i| i.even };  
		synthCtrs = this.synthCtr.select {|x,i| i.even };  
		
		case
			{ synth.isNil }
				{ 
					(this.synthCtr.size != 0).if {   
						SimpleInitError("synthCtr input but no synth input. " ++
							"Synth must be Symbol, String (indicating a Synthdef known by SynthDesLib), " ++ 
								"Synth or Integer (nodeID) or collection thereof.").class_(VarGui).throw;
					}{
						synths = [];
					};
				}
			{ synth.isKindOf(SequenceableCollection) }
				{ 
					synths = synth;
				}
			{ [Synth, Integer, Symbol, String].any(synth.isKindOf(_)) }
				{ 
					synths = [synth];
				}
			{ true }{ SimpleInitError("Synth must be Symbol, String (indicating a Synthdef known by SynthDesLib), " ++ 
				"Synth or Integer (nodeID) or collection thereof.").class_(VarGui).throw; };
						
		
		synths.do {|x| this.checkSingleSynthInput(x) };
		
		synthStateVars = [assumePlaying, assumeRunning, assumeEnded];
		synthStateStrings = ["assumePlaying", "assumeRunning", "assumeEnded"];
		
		synthStateStrings.do {|x,i|
			case
				{ synthStateVars[i].isKindOf(SequenceableCollection) }
					{ 
						(synthStateVars[i].every {|x| [Boolean, Nil].any(x.isKindOf(_)) } && 
							(synthStateVars[i].size == synths.size)).not.if {   
								SimpleInitError(x ++ " input as collection must equal number of synths, items must be Boolean or Nil").class_(VarGui).throw;
						}{
							this.perform((x ++ "_").asSymbol, synthStateVars[i]);
						}
					}
				{ [Boolean, Nil].any(synthStateVars[i].isKindOf(_)) }
					{ 
						this.perform((x ++ "_").asSymbol, synthStateVars[i].dup(synths.size));
					}
				{ true }{ SimpleInitError(x ++ " must be Boolean, Nil or collection thereof, " ++
							"in this case size must corespond with synthCtr").class_(VarGui).throw; };
		};
		
		this.assumeEnded.do {|x,i|
			((x == true) && ((this.assumePlaying[i] == true) || (this.assumeRunning[i] == true))).if {
				SimpleInitError("Contradictory assumptions about synth state, there must not be a player with assumeEnded set true " ++
					"together with assumePlaying or " ++ "assumeRunning set true").class_(VarGui).throw;
			};
			((this.assumePlaying[i] == false) && (this.assumeRunning[i] == true)).if {
				SimpleInitError("Contradictory assumptions about synth state, there must not be a player with assumePlaying set false " ++
					"and assumeRunning set true").class_(VarGui).throw;
			};		
		};
				
		synthStateStrings.do {|x,i|
			this.perform(x.asSymbol).do {|y,j|
				(([Symbol, String].any(synths[j].isKindOf(_))) && (y.notNil)).if {
					SimpleInitError("There must not be any assume flag set to a Boolean for a Symbol / String as synth input, " ++
						"assumptions are only allowed for Synths and nodeIDs.").class_(VarGui).throw;
				}
			}
		}; 
			
		
		synthIDs = synths.collect {|x| [Synth, Integer].any(x.isKindOf(_)).if { x.asNodeID } };
		
		stream = stream.asArray;
		streamEnvirs = List.new;
		streamEnvirIndices = Array.newClear(stream.size);
		streams = Array.newClear(stream.size);

		stream.do {|x, i|
			var e, k;
			case
				{ x.isKindOf(Function) }{ e = Environment.new; streams[i] = e.use { Task(x) } }
				{ x.isKindOf(Pattern) }{ e = Environment.new; streams[i] = e.use { x.asEventStreamPlayer } }
				{ true }{
					x.isKindOf(PauseStream).not.if {
						SimpleInitError("Stream must be Task, Task function, Pattern, EventStreamPlayer or collection of such").class_(VarGui).throw;
					}{
						e = x.originalStream.miSC_getEnvironment;
						k = streamEnvirs.detectIndex(_ === e);
						streams[i] = x;
					}
				};
			k.notNil.if {
				streamEnvirIndices[i] = k;
			}{
				streamEnvirs = streamEnvirs.add(e);
				streamEnvirIndices[i] = j;
				j = j + 1;
			};
		};
		
		envir.isNil.if {
			envirs = (stream.size == 0).if { [currentEnvironment] }{ streamEnvirs };
		}{
			(stream.size != 0).if { SimpleInitError("Passing an envir arg only allowed without stream players").class_(VarGui).throw };
			envirs = case
				{ envir.isKindOf(Environment) }{ [envir] }
				{ envir.isSequenceableCollection }{ envir }
				{ true }{ SimpleInitError("Envir arg  must be Environment or SequenceableCollection").class_(VarGui).throw; };
			
			envirSet = IdentitySet.new;
			envirs.do { |envir| envirSet = envirSet.add(envir)};
			
			((envirs.any(_.isKindOf(Environment).not)) || (envirSet.size != envirs.size)).if {
				SimpleInitError("Envir arg must be Environment or SequenceableCollection of non-identical Environments").class_(VarGui).throw;
			};
		
		};

		varCtr.first.isSequenceableCollection.if {
			varEnvirGroups.notNil.if {
				SimpleInitError("varEnvirGroups must not be given if varCtr is already grouped, means a collection of specPair collections").class_(VarGui).throw;
			};
			hasConnectedVarEnvirGroups = true;
			envir.isNil.if {
				(stream.size != 0).if {
					((varCtr.size) > (streamEnvirs.size)).if {
						SimpleInitError("Size of varCtr exceeds number of implicitely given stream envirs").class_(VarGui).throw;
					}
				}{
					((varCtr.size) > 1).if {
						SimpleInitError("Size of varCtr > 1 but no implicitely or explicitely given envirs").class_(VarGui).throw;
					}
				}
			}{	// can only happen if no streams 
				((varCtr.size) > (envirs.size)).if {
					SimpleInitError("Size of varCtr exceeds number of explicitely given envirs").class_(VarGui).throw;
				};
			};
			varCtr.do { |specPairs, i|
				var envirKeys; 
				specPairs.do { |item, j| 
					j.even.if { 
						varEnvirIndices = varEnvirIndices.add(i);
						envirKeys = envirKeys.add(item); 
					} 
				};
				(envirKeys.asSet.size != envirKeys.size).if {
					SimpleInitError("Bad definition: specPairs defined for one envir contain var name more than once").class_(VarGui).throw;
				} 
			};
		}{
			varEnvirGroups.isNil.if {
				(vars.size != vars.asSet.size).if {
					SimpleInitError("varCtr contains multiple keys but varEnvirGroups not defined").class_(VarGui).throw; 
				};
				hasConnectedVarEnvirGroups = true;
				varEnvirIndices = vars.size.collect { 0 };
			}{
				varEnvirGroups.miSC_isGrouping(vars.size).not.if {
					SimpleInitError("varEnvirGroups must be a valid grouping of number of vars. E.g. [[3,0], [1,2,4]] is a valid grouping of 5").class_(VarGui).throw; 
				};
				hasConnectedVarEnvirGroups = varEnvirGroups.miSC_isConnected;
				(varEnvirGroups.size > envirs.size).if {
					SimpleInitError("Size of varEnvirGroups must less or equal " ++ envirs.size.asString ++ ", the number of " ++ 
						(envir.notNil.if { "ex" }{ "im" }) ++  "plicitely given envirs").class_(VarGui).throw; 
				};
				varEnvirIndices = Array.newClear(vars.size);
				varEnvirGroups.do { |x,i|
					var y = vars[x];
					(y.size != y.asSet.size).if { 
						SimpleInitError("Bad definition: envirGroup " ++ x.asString ++ " contains var name more than once").class_(VarGui).throw;
					};
					x.do(varEnvirIndices[_] = i);
				};
			};
		};				
		
				
		synthCtr.first.isSequenceableCollection.if {
			synthCtrGroups.notNil.if {
				SimpleInitError("synthCtrGroups must not be given if synthCtr is already grouped, means a collection of specPair collections").class_(VarGui).throw;
			};
			hasConnectedSynthCtrGroups = true;
				
			(synthCtr.size > synths.size).if {
				SimpleInitError("Size of synthCtr exceeds number of given synths").class_(VarGui).throw;
			};
			synthCtr.do { |specPairs, i|
				var synthKeys; 
				specPairs.do { |item, j| 
					j.even.if { 
						synthCtrSynthIndices = synthCtrSynthIndices.add(i);
						synthKeys = synthKeys.add(item); 
					} 
				};
				(synthKeys.asSet.size != synthKeys.size).if {
					SimpleInitError("Bad definition: specPairs defined for one synth contain ctr name more than once").class_(VarGui).throw;
				} 
			};
		}{
			synthCtrGroups.isNil.if {
				(synthCtrs.size != synthCtrs.asSet.size).if {
					SimpleInitError("synthCtr contains multiple keys but synthCtrGroups not defined").class_(VarGui).throw; 
				};
				hasConnectedSynthCtrGroups = true;
				synthCtrSynthIndices = synthCtrs.size.collect { 0 };
			}{
				synthCtrGroups.miSC_isGrouping(synthCtrs.size).not.if {
					SimpleInitError("synthCtrGroups must be a valid grouping of number of synthCtrs. E.g. [[3,0], [1,2,4]] is a valid grouping of 5").class_(VarGui).throw; 
				};
				hasConnectedSynthCtrGroups = synthCtrGroups.miSC_isConnected;
				(synthCtrGroups.size > synths.size).if {
					SimpleInitError("Size of synthCtrGroups must be less or equal " ++ synths.size.asString ++ ", the number of given synths").class_(VarGui).throw; 
				};
				synthCtrSynthIndices = Array.newClear(synthCtrs.size);
				synthCtrGroups.do { |x,i|
					var y = synthCtrs[x];
					(y.size != y.asSet.size).if { 
						SimpleInitError("Bad definition: synthCtrGroup " ++ x.asString ++ " contains synth controlname more than once").class_(VarGui).throw;
					};
					x.do(synthCtrSynthIndices[_] = i);
				};
			};
		};				
	
	hasConnectedSynthCtrGroups.if { 
		firstSynthCtrIndices = (0..(synthCtrSynthIndices.size - 1)).select { |i| 
			((i == 0) or: { (synthCtrSynthIndices[i-1] != synthCtrSynthIndices[i] ) }) 
		}
	};
	
	
	(envir.notNil && (varEnvirIndices.asSet.size < envirs.size)).if { 
		SimpleInitError("Bad definition: unused explicitely given envirs").class_(VarGui).throw;
	};
	
	varColorNum = varEnvirIndices.asSet.size;
	synthColorNum = synths.size;
	
	^this;	
	}
}

